from contractme import contracting

ignore_preconditions = contracting.ignore_preconditions
check_preconditions = contracting.check_preconditions
ignore_postconditions = contracting.ignore_postconditions
check_postconditions = contracting.check_postconditions

precondition = contracting.precondition
postcondition = contracting.postcondition
annotated = contracting.annotated
